package wpd

import (
	"bufio"
	"context"
	"encoding/json"
	"fmt"
	"io"
	"sync"
	"sync/atomic"
	"time"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"
)

// Runner ...
type Runner interface {
	Stdin() io.WriteCloser
	Stdout() io.ReadCloser
	Stop() error
}

// RunnerInfo ...
type RunnerInfo struct {
	Wpd      string `json:"wpd"`
	Version  string `json:"version"`
	Protocol string `json:"protocol"`
}

// RunnerFactory is a function that creates a Runner
type RunnerFactory func() (Runner, error)

// Job ...
type Job struct {
	Docname string       `json:"-"`
	ID      uint64       `json:"id"`
	CSS     string       `json:"css"`
	HTML    []byte       `json:"-"`
	Size    int          `json:"size"`
	Result  chan *Result `json:"-"`
	started time.Time
	context context.Context
	stop    chan struct{}
}

func (job *Job) success(r *Result) {
	close(job.stop)
	renderDocumentDuration.WithLabelValues(
		job.Docname, "success",
	).Observe(float64(time.Since(job.started)) / float64(time.Millisecond))
	job.Result <- r
}

func (job *Job) failure(err error) {
	close(job.stop)
	renderDocumentDuration.WithLabelValues(
		job.Docname, "success",
	).Observe(float64(time.Since(job.started)) / float64(time.Millisecond))
	job.Result <- &Result{ID: job.ID, Error: err}
}

// Result ...
type Result struct {
	ID      uint64 `json:"id"`
	Content []byte `json:"-"`
	Size    int    `json:"size"`
	Error   error  `json:"error"`
}

// WPD ...
type WPD struct {
	wg      sync.WaitGroup
	queue   chan *Job
	results chan *Result

	runner Runner

	counter uint64

	quit chan struct{}

	log zerolog.Logger
}

// WriteAll writes a list of []byte to a writer in one shot
func WriteAll(w io.Writer, data ...[]byte) (int, error) {
	var n int
	for _, d := range data {
		nn, err := w.Write(d)
		n += nn
		if err != nil {
			return n, err
		}
	}
	return n, nil
}

func heloRunner(runner Runner, reader *bufio.Reader, log zerolog.Logger) (err error) {
	t := time.AfterFunc(time.Second*5, func() {
		if err := runner.Stop(); err != nil {
			log.Err(err).Msg("Error stopping the runner")
		}
	})
	defer func() {
		if !t.Stop() {
			err = fmt.Errorf("timeout reading runner info header")
		}
	}()

	line, err := reader.ReadBytes(byte('\n'))
	if err != nil {
		fmt.Println(err)
		return err
	}
	var info RunnerInfo
	if err := json.Unmarshal(line, &info); err != nil {
		panic(err)
	}
	if info.Wpd == "" {
		return fmt.Errorf("invalid wpd info: missing \"wpd\"")
	}
	if info.Protocol != "1" {
		return fmt.Errorf("invalid wpd protocol version. Expects \"1\", got: \"%s\"", info.Protocol)
	}
	return nil
}

// NewWPD ...
func NewWPD(runnerFactory RunnerFactory, log zerolog.Logger) (*WPD, error) {
	runner, err := runnerFactory()
	if err != nil {
		return nil, err
	}
	reader := bufio.NewReader(runner.Stdout())
	if err := heloRunner(runner, reader, log); err != nil {
		if err := runner.Stop(); err != nil {
			log.Err(err).Msg("Error stopping the runner")
		}
		return nil, err
	}

	var wpd = WPD{
		queue:   make(chan *Job),
		results: make(chan *Result),
		runner:  runner,
		quit:    make(chan struct{}),
		log:     log,
	}

	wpd.wg.Add(2)
	go wpd.readWpdOutput(reader)
	go wpd.run()

	return &wpd, nil
}

// Render ...
func (wpd *WPD) Render(ctx context.Context, docname, css string, html []byte) ([]byte, error) {
	job := wpd.Submit(ctx, docname, css, html)
	result := <-job.Result
	return result.Content, result.Error
}

// Submit ...
func (wpd *WPD) Submit(ctx context.Context, docname, css string, html []byte) *Job {
	id := atomic.AddUint64(&wpd.counter, 1)

	var job = Job{
		Docname: docname,
		ID:      id,
		CSS:     css,
		HTML:    html,
		Size:    len(html),
		Result:  make(chan *Result, 1),
		started: time.Now(),
		context: ctx,
		stop:    make(chan struct{}),
	}
	wpd.queue <- &job
	return &job
}

// Close ..
func (wpd *WPD) Close() error {
	close(wpd.quit)
	// now close the runner
	runnerStopErr := wpd.runner.Stop()
	wpd.wg.Wait()
	close(wpd.queue)
	close(wpd.results)
	for range wpd.queue {
	}
	for range wpd.results {
	}
	return runnerStopErr
}

func (wpd *WPD) run() {
	defer wpd.wg.Done()
	defer wpd.runner.Stdin().Close()

	var (
		running      = make(map[uint64]*Job)
		canceledJobs = make(chan *Job)
	)

	defer func() {
		for _, job := range running {
			job.failure(fmt.Errorf("WPD is shutting down"))
		}
	}()

	for {
		select {
		case <-wpd.quit:
			return
		case job := <-wpd.queue:
			if b, err := json.Marshal(job); err != nil {
				job.failure(err)
			} else {
				if _, err := WriteAll(
					wpd.runner.Stdin(), b, []byte("\n"), job.HTML, []byte("\n"),
				); err != nil {
					// TODO handle write errors
					job.failure(err)
				} else {
					running[job.ID] = job
					go func() {
						select {
						case <-wpd.quit:
							return
						case <-job.stop:
							return
						case <-job.context.Done():
							canceledJobs <- job
						}
					}()
				}
			}
		case result := <-wpd.results:
			if job, ok := running[result.ID]; ok {
				job.success(result)
				delete(running, result.ID)
			}
		case job := <-canceledJobs:
			if job, ok := running[job.ID]; ok {
				delete(running, job.ID)
				job.failure(job.context.Err())

				b, err := json.Marshal(map[string]interface{}{
					"id":     job.ID,
					"action": "cancel",
				})
				if err != nil {
					panic(err)
				}
				if _, err := WriteAll(
					wpd.runner.Stdin(), b, []byte("\n"),
				); err != nil {
					// TODO handle write errors
					wpd.log.Err(err).Msg("could not write to wpd")
				}
			}
		}
	}
}

func (wpd *WPD) readWpdOutput(reader *bufio.Reader) {
	defer wpd.wg.Done()

	for {
		line, err := reader.ReadBytes(byte('\n'))
		if err != nil {
			return
		}
		var r Result
		if err := json.Unmarshal(line, &r); err != nil {
			panic(err)
		}
		if r.Size != 0 {
			var buf = make([]byte, r.Size)
			_, err := io.ReadFull(reader, buf)
			if err != nil {
				panic(err)
			}

			if b, err := reader.ReadByte(); err != nil || b != byte('\n') {
				panic(err)
			}
			r.Content = buf
		}
		wpd.results <- &r
	}
}

var (
	renderDocumentDuration = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "alpe_wpd_render",
			Help:       "The WPD rendering duration and count",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
		},
		[]string{"docname", "status"},
	)
)
