package tests

import (
	"bytes"
	"net/http"
	"testing"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"orus.io/orus-io/rednerd/models"
)

func TestAuthentication(t *testing.T) {
	tester := NewAPITester(t)
	defer tester.Close()

	tester.Run("WebToken", func(t *testing.T) {
		defer tester.SetHeader("Authorization", "")

		var token string
		require.Equal(t, 200, tester.MustPostJSON("/api/v1/auth/authenticate", JSONObj{
			"username": "admin",
			"password": "admin",
		}, &token))

		tester.SetHeader("Authorization", "Bearer "+token)

		var userinfo models.Userinfo
		require.Equal(t, 200, tester.MustGetJSON("/api/v1/auth/userinfo", &userinfo))

		assert.Equal(t, "admin", userinfo.Username)
		assert.True(t, userinfo.IsAdmin)
	})

	tester.Run("Cookie", func(t *testing.T) {
		defer tester.SetHeader("Cookie", "")

		req, _ := http.NewRequest("POST", "/api/v1/auth/authenticate",
			bytes.NewBufferString(`{"username":"admin","password":"admin","set-cookie":true}`))
		req.Header.Set("Content-Type", "application/json")

		res := tester.HandleRequest(req)
		t.Log(res.Header().Get("Set-Cookie"))
		for _, cookie := range res.Result().Cookies() {
			require.Equal(t, "rednerd-auth", cookie.Name)
			tester.SetHeader("Cookie", cookie.String())
		}

		var userinfo models.Userinfo
		require.Equal(t, 200, tester.MustGetJSON("/api/v1/auth/userinfo", &userinfo))

		assert.Equal(t, "admin", userinfo.Username)
		assert.True(t, userinfo.IsAdmin)
	})

	tester.Run("APIKey", func(t *testing.T) {
		defer tester.Logout()
		defer tester.SetHeader("Rednerd-API-Key", "")

		tester.Login("admin", "admin")
		require.Equal(t, http.StatusCreated,
			tester.MustPostJSON("/api/v1/user", JSONObj{
				"username": "janedoe",
			}, &JSONObj{}))
		var apikey models.Apikey
		require.Equal(t, http.StatusCreated,
			tester.MustPostJSON("/api/v1/user/janedoe/apikey", JSONObj{
				"name": "key1",
			}, &apikey))
		tester.Logout()

		tester.SetHeader("Rednerd-API-Key", apikey.Value)

		var userinfo models.Userinfo
		require.Equal(t, http.StatusOK,
			tester.MustGetJSON("/api/v1/auth/userinfo", &userinfo))
		assert.Equal(t, "janedoe", userinfo.Username)
		assert.False(t, userinfo.IsAdmin)
	})
}
